#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <stdio.h>
#include <stdlib.h>
#include <util/delay.h>
#include <string.h>
#include <math.h>
#include "uart.h"
#define WAITING 14
#define HITCHECK 15
#define HITWAIT 16
#define NOTHITCHECK 17
#define countTwoFifthMS 25  //ticks/mSec


FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);


int programSel;
int digVolt;
int sovOld;
char debounceState;
volatile unsigned long reax = 0 ;
volatile unsigned long time = 0 ;
volatile unsigned long timeGlobal = 0;
double timeDebounce;
char count;



/*
ISR (TIMER3_OVF_vect) {
	time_ovfl = time_ovfl + 65536;
	time++;
	timeGlobal++;
		//pollTimer++;
        //PORTB ^= 2;
	//if(time_ovfl > 300000000000){ //30 bil = 1500 seconds
	//		TCNT3 = 0;
	//		PORTB ^= 2;
//	}
}

*/
ISR (TIMER0_OVF_vect) {
	reax++;
	time++;
	timeGlobal++;
}



void main(void) {
	//various programs, with timings 
	char lenEasy = 25;
	int progEasy[25] = {1, 1, 3, 4, 3, 2, 5, 2, 1, 1, 3, 2, 3, 4, 1, 1, 3, 2, 3, 2, 5, 2, 1, 1, 3}; 
	int  releaseEasy[25] = {250, 500, 400, 600, 400, 400, 700, 400, 200, 500, 200, 400, 700, 250, 500, 400, 600, 400, 400, 250, 500, 400, 600, 400, 400};
	int  timingEasy[25] = {1250, 2500, 2400, 3000, 2500, 4000, 3000, 1400, 2000, 2500, 1000, 2000, 4000, 1250, 2500, 2400, 3000, 2500, 4000, 1250, 2500, 2400, 3000, 2500, 4000  };
	char lenMed = 7;
	int progMed[7] = {1, 2, 2, 1, 1, 2, 2};
	int  timingMed[7] = {2700, 2700, 2700, 2700, 2700, 2700, 2700};
	int  releaseMed[7] = {1700, 1700, 1700, 1700, 1700, 1700, 1700};
	char lenHard = 5;
	int progHard[5] = {1, 2, 3, 4, 5};
	int  timingHard[5] = {8000, 4000, 12000, 6000, 3000};
	int  releaseHard[5] = {8000, 4000, 12000, 6000, 3000};

	//Pointers for paramteres
	int* timingList;
	int* punchList;
	int* releaseList;

	//Self program array initializations
	int progSelf[100] = {0};
	int releaseSelf[100] = {0};
	int timingSelf[100] = {0};
	char didHitList[100] = {0};


	double reaxList[100] = {0};
	double rOne=0;
	double rTwo=0;
	double rThree=0;
	double rFour=0;
	double rFive=0;
	float accOne =0; 
	float accTwo =0; 
	float accThree = 0;
	float accFour = 0;
	float accFive = 0;
	float nOne = 0;
	float nTwo=0;
	float nThree=0;
	float nFour=0;
	float nFive=0;
	char lenProgram=0;

	char finished;
	char pos;
	char runIter = 0;;
	int yesOrNo;
	int i;
	uart_init();
	stdout = stdin = stderr = &uart_str;
	programSel = 0;
	for(i=0;i<100;i++){
		releaseSelf[i] = 100;
	}

	sei();
	ADMUX = (1<<ADLAR) | (1<<REFS0)  ; //5 V reference, p 259 for selector
	ADCSRA = ((1<<ADEN) | (1<<ADSC)) + 6 ; // Last three bits are clock prescalar, pg. 260. prescalar = 2^number, 000 = 2x 
 	DDRB = 0xff;
	TCCR0B = 1 ;  //Timer0 prescalar: 1
	TIMSK0 = (1<<TOIE0) ; //Enable ovf interrupt
	while(1){
		if(runIter == 0) fprintf(stdout,"Welcome to Punch Punch Revolution\n\r"); 
		fprintf(stdout, "Please select your desired program number, use ENTER to confirm \n\r");
		fprintf(stdout, "[1] Easy [2] Medium [3] Hard [4] Self-Programmed \n\r");
		fscanf(stdin, "%d", &programSel);
		//If block to set program
		if (programSel==1) {
			punchList = progEasy;
			timingList = timingEasy;
			releaseList = releaseEasy;
			lenProgram = lenEasy;
		}
		else if (programSel == 2) {
			punchList = progMed;
			timingList = timingMed;
			releaseList = releaseMed;
			lenProgram = lenMed;
		}
		else if (programSel == 3) {
			punchList = progHard;
			timingList = timingHard;
			releaseList = releaseHard;
			lenProgram = lenHard;
		}
		else if (programSel == 4) { //Self programmed
			fprintf(stdout, "Load a prior program with 1, create a new one with 2");
			fscanf(stdin, "%d", &yesOrNo);
				if(yesOrNo==1){
					yesOrNo = 2;
					lenProgram=eeprom_read_byte((uint8_t*)0); //Define length of program and place at beginning of eeprom
					eeprom_read_block(punchList, (uint8_t*)1, lenProgram);
					eeprom_read_block(releaseList, (uint16_t*)(lenProgram+1), 2*lenProgram);
					eeprom_read_block(timingList, (uint16_t*)(3*lenProgram+1), 2*lenProgram);
				}
				else{
					fprintf(stdout,"Program your own practice routine here \n\r");
					fprintf(stdout,"First enter the number of the section to hit \n\r");
					fprintf(stdout,"Type 9 to finish the sequence \n\r");
					fprintf(stdout,"Then add a timing restriction, up to 65000ms \n\r");
					fprintf(stdout,"The lay out is shown below: \n\r");
					fprintf(stdout,"  [1]\n\r");
					fprintf(stdout,"[2] [3] \n\r");
					fprintf(stdout,"[4] [5] \n\r");
					for (i=0; i<200; i++) {
						fprintf(stdout, "Section for punch %i: \n", i+1);
						fscanf(stdin, "%d", &progSelf[i]);
						if (progSelf[i] == 9) {
							lenProgram = i;
							punchList = progSelf;
							timingList = timingSelf;
							releaseList = releaseSelf;
							fprintf(stdout, "Would you like to save your program to memory? Yes=[1], No=[2]? \n\r WARNING: This will delete any previous program \n\r");
							fscanf(stdin, "%d", &yesOrNo);
							if(yesOrNo==1){
								yesOrNo = 2;
								eeprom_write_byte((uint8_t*)0, lenProgram); //Define length of program and place at beginning of eeprom
								eeprom_write_block(punchList, (uint8_t*)1, lenProgram);
								eeprom_write_block(releaseList, (uint16_t*)(lenProgram+1), 2*lenProgram);
								eeprom_write_block(timingList, (uint16_t*)(3*lenProgram+1), 2*lenProgram);
							}
							break;
						}
						fprintf(stdout, "Delay from previous punch %i: ms\n", i+1);
						fscanf(stdin, "%d", &releaseSelf[i]);
						fprintf(stdout, "Deadline, relative from delay, for punch %i: ms\n", i+1);
						fscanf(stdin, "%d", &timingSelf[i]);	
					}
				}
		}
		
		fprintf(stdout, "Program Length is %u punches \n\r", lenProgram );
		for (i=0; i<lenProgram; i++){
			fprintf(stdout, "Punch: %d Release: %d Deadline: %d \n\r", punchList[i], releaseList[i], releaseList[i]+timingList[i]);
		}
		timeGlobal=0;
		for (i=0; i<lenProgram; i++){ //Generate punches and time them
				sovOld = 0;
				time = 0;
				while(time<(releaseList[i]*62.5));
				finished = 0;
				//time_ovfl = 0;
				reax=0;
				TCNT0 = 0;
				debounceState = WAITING;
				pos = punchList[i];
				ADMUX = (1<<ADLAR) | (1<<REFS0); //Set up ref voltage
				switch(pos){ //Select ADC and trigger LED
					case 1: 
						PORTB = 1; 
						break;
					case 2: 
						PORTB = 2; 
						ADMUX += 1;
						break;
					case 3:
						PORTB = 4; 
						ADMUX +=2;
						break;
					case 4:
						PORTB = 8;
						ADMUX +=3;
						break; 
					case 5:
						PORTB = 16;
						ADMUX +=4;
						break; 
					default:
						PORTB = 0;
				}
				do{
					ADCSRA |=(1<<ADSC);	
					while(ADCSRA&(1<<ADSC)); //is conversion done?		
					digVolt = ADCH;
						switch(debounceState){
						case WAITING:
							if (digVolt>25){
								sovOld = digVolt;
								reaxList[i] = (float)(reax + (TCNT0/256))/62.5;
								didHitList[i] = 1;		
								debounceState = HITWAIT;
							}
							break;
						case HITWAIT:
							PORTB = 0x00;
							if (digVolt<20){debounceState = NOTHITCHECK;}
							else if (digVolt > sovOld) {
								sovOld = digVolt;
							} 
							break;
						case NOTHITCHECK:
							if (digVolt<20){
								debounceState = WAITING;
								finished = 1;
							} 
							break; 
						}				
				}while(!finished && (time<((timingList[i]+releaseList[i])*62.5)));//replace true with timeout
				if(didHitList[i]) {
					fprintf(stdout, "Punch%d - React:%.1fms Time:%.3fs  SPI:%d \n\r", i+1, reaxList[i], timeGlobal/62500.0, sovOld);	//SPI = Soviet Punching Index
				}
				else {
					fprintf(stdout, "Punch%d - MISSED. You had %d ms to hit. Time:%.3fs \n\r", i+1, timingList[i], timeGlobal/62500.0);
				}			
		}
		PORTB = 0;
		for (i=0; i<lenProgram; i++){
			if(punchList[i]==1){
				nOne++;
				accOne = accOne + didHitList[i];
				rOne = rOne + reaxList[i];
			}
			else if(punchList[i]==2){
				nTwo++;
				accTwo = accTwo + didHitList[i];
				rTwo = rTwo + reaxList[i];
			}
			else if(punchList[i]==3){
				nThree++;
				accThree = accThree + didHitList[i];
				rThree = rThree + reaxList[i];
			}
			else if(punchList[i]==4){
				nFour++;
				accFour = accFour + didHitList[i];
				rFour = rFour + reaxList[i];
			}
			else if(punchList[i]==5){
				nFive++;
				accFive = accFive + didHitList[i];
				rFive = rFive + reaxList[i];
			}
		} 
		accOne = accOne/nOne;
		accTwo = accTwo/nTwo;
		accThree = accThree/nThree;
		accFour = accFour/nFour;
		accFive = accFive/nFive;
		rOne = rOne/nOne;
		rTwo = rTwo/nTwo;
		rThree = rThree/nThree;
		rFour = rFour/nFour;
		rFive = rFive/nFive;
		fprintf(stdout,"      [ %5.3f ]\n\r", accOne);
		fprintf(stdout,"      [ %5.3f ]\n\r", rOne);
		fprintf(stdout,"  [ %5.3f || %5.3f ]\n\r", accTwo, accThree);
		fprintf(stdout,"  [ %5.3f || %5.3f ]\n\r", rTwo, rThree);
		fprintf(stdout,"  [ %5.3f || %5.3f ]\n\r", accFour, accFive);
		fprintf(stdout,"  [ %5.3f || %5.3f ]\n\r", rFour, rFive);
		fprintf(stdout, "Would you like to play again? Yes=[1], No=[2]? \n\r ");
		fscanf(stdin, "%d", &yesOrNo);
		if(yesOrNo==1){
			runIter++;
			yesOrNo = 2;
		}
		else break;
	}
}
